/*
 * Mach Operating System
 * Copyright (c) 1992 Carnegie Mellon University
 * All Rights Reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie Mellon
 * the rights to redistribute these changes.
 */
/* This is a rewrite (retype) of the Amiga's custom chip register map, based
   on the Hardware Reference Manual.  It is NOT based on the Amiga's
   hardware/custom.h.  */

#ifndef _machine_custom_
#define _machine_custom_

#include <machine/vm_param.h>

#include "i:hardware/intbits.h"
#include "i:hardware/dmabits.h"

#define CUSTOM_P 0xdff000
/*
#define CUSTOM_V (0-PAGE_SIZE+(CUSTOM_P & (PAGE_SIZE-1)))
*/
#define CUSTOM_V 0x4ffff000

struct Custom
  {
    /*** read-only registers ***/

    short zz1;
    short dmaconr;
    short vposr;
    short vhposr;
    short zz2;
    short joy0dat;
    short joy1dat;
    short clxdat;
    short adkconr;
    short pot0dat;
    short pot1dat;
    short potgor;
    short serdatr;
    short dskbytr;
    short intenar;
    short intreqr;

    /*** write-only registers ***/

    /* disk */
    void *dskpt;
    short dsklen;

    short zz3[2];
    short vposw;
    short vhposw;
    short copcon;
    short serdat;
    short serper;
    short potgo;
    short joytest;
    short zz4[4];

    /* blitter */
    short bltcon0;
    short bltcon1;
    short bltafwm;
    short bltalwn;
    void *bltcpt;
    void *bltbpt;
    void *bltapt;
    void *bltdpt;
    short bltsize;
    short zz5[3];
    short bltcmod;
    short bltbmod;
    short bltamod;
    short bltdmod;
    short zz6[4];
    short bltcdat;
    short bltbdat;
    short bltadat;
    short zz7[4];

    /* more disk */
    short dsksync;

    /* copper */
    void *cop1lc;
    void *cop2lc;
    short copjmp1;
    short copjmp2;
    short copins;

    /* display parameters */
    short diwstrt;
    short diwstop;
    short ddfstrt;
    short ddfstop;

    /* control registers */
    short dmacon;
    short clxcon;
    short intena;
    short intreq;

    /* audio */
    short adkcon;
    struct Audio
      {
        void *lc;
        short len;
        short per;
        short vol;
        short zz[3];
      } aud[4];

    /* display */
    void *bplpt[6];
    short zz8[4];
    short bplcon0;
    short bplcon1;
    short bplcon2;
    short zz9;
    short bpl1mod;
    short bpl2mod;
    short zz10[2+6+2];

    /* sprites */
    void *sprpt[8];
    struct Sprite
      {
        short pos;
        short ctl;
        short data;
        short datb;
      } spr[8];

    short color[32];
  };

#define custom (*((volatile struct Custom *)CUSTOM_V))

#endif /* _machine_custom_ */
