/*
 * Copyright (c) 1988 University of Utah.
 * Copyright (c) 1982, 1990 The Regents of the University of California.
 * All rights reserved.
 *
 * This code is derived from software contributed to Berkeley by
 * the Systems Programming Group of the University of Utah Computer
 * Science Department.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * from: Utah $Hdr: frame.h 1.1 90/07/09$
 *
 *      @(#)frame.h     7.2 (Berkeley) 11/2/90
 */
#ifndef _machine_frame_
#define _machine_frame_

#include <machine/trap.h>

/*
 * User registers are pushed on PCB stack.  Bus error frame
 * follows this.  Variable length - the end address is fixed.
 *
 * Kernel registers are pushed on kernel stack.
 */
struct mc68020_saved_state {
};


/* FIXME: These should probably be somewhere else... */
#define SR_TRACE_MASK   0xc000
#define SR_T1           0x8000
#define SR_T0           0x4000
#define SR_SUPERVISOR   0x2000
#define SR_MASTER       0x1000
#define SR_INT_MASK     0x0700
#define SR_INT_SHIFT    8


/*
 * Other user machine state, saved on context switch.
 */

struct frame {
        int             sp;             /* user stack pointer */

        int             d0;             /* data registers */
        int             d1;
        int             d2;
        int             d3;
        int             d4;
        int             d5;
        int             d6;
        int             d7;
        int             a0;             /* address registers */
        int             a1;
        int             a2;
        int             a3;
        int             a4;
        int             a5;
        int             a6;

        unsigned short  sr;             /* status register */
        unsigned int    pc;             /* program counter - UNALIGNED!!! */
        unsigned int    format  : 4;    /* rte stack frame format */
        unsigned int    vector  : 12;   /* vector number */

        union F_u {
                struct fmt2 {
                        unsigned int    iaddr;
                } fmt2;

                struct fmt9 {
                        unsigned int    iaddr;
                        unsigned short  iregs[4];
                } fmt9;

                struct fmtA {
                        unsigned short  ir0;

                        unsigned short  faultb : 1;
                        unsigned short  faultc : 1;
                        unsigned short  rerunb : 1;
                        unsigned short  rerunc : 1;
                        unsigned short  unused_1 : 3;
                        unsigned short  dfault : 1;
                        unsigned short  rmw : 1;
                        unsigned short  rw : 1;
                        unsigned short  size : 2;
                        unsigned short  unused_2 : 1;
                        unsigned short  fc : 3;

                        unsigned short  ipsc;
                        unsigned short  ipsb;
                        unsigned int    dfa;
                        unsigned short  ir1, f_ir2;
                        unsigned int    dob;
                        unsigned short  ir3, f_ir4;
                } fmtA;

                struct fmtB {
                        unsigned short  ir0;

                        unsigned short  faultb : 1;
                        unsigned short  faultc : 1;
                        unsigned short  rerunb : 1;
                        unsigned short  rerunc : 1;
                        unsigned short  unused_1 : 3;
                        unsigned short  dfault : 1;
                        unsigned short  rmw : 1;
                        unsigned short  rw : 1;
                        unsigned short  size : 2;
                        unsigned short  unused_2 : 1;
                        unsigned short  fc : 3;

                        unsigned short  ipsc;
                        unsigned short  ipsb;
                        unsigned int    dfa;
                        unsigned short  ir1, f_ir2;
                        unsigned int    dob;
                        unsigned short  ir3, f_ir4;
                        unsigned short  ir5, f_ir6;
                        unsigned int    sba;
                        unsigned short  ir7, f_ir8;
                        unsigned int    dib;
                        unsigned short  iregs[22];
                } fmtB;
        } u;
};

struct kern_frame {
        int             d0;
        int             d1;
        int             a0;
        int             a1;
        unsigned short  sr;             /* status register */
        unsigned int    pc;             /* program counter - UNALIGNED!!! */
        unsigned int    format  : 4;    /* rte stack frame format */
        unsigned int    vector  : 12;   /* vector number */
        union F_u u;
};

#define FMT0            0x0
#define FMT1            0x1
#define FMT2            0x2
#define FMT9            0x9
#define FMTA            0xA
#define FMTB            0xB

/* frame specific info sizes */
#define FMT0SIZE        0
#define FMT1SIZE        0
#define FMT2SIZE        sizeof(struct fmt2)
#define FMT9SIZE        sizeof(struct fmt9)
#define FMTASIZE        sizeof(struct fmtA)
#define FMTBSIZE        sizeof(struct fmtB)

#define V_BUSERR        (T_BUSERR*4)
#define V_ADDRERR       (T_ADDRERR*4)
#define V_TRACE         (T_TRACE*4)
#define V_TRAP1         0x084

struct fp_status
  {
    long status;
    long fpcr;                                  /* FP control registers */
    long fpsr;
    long fpiar;
    long regs[8][3];                            /* 96-bit general purpose FP registers */
  };

#define FPS_UNUSED      0                       /* Thread hasn't used any FP instructions (yet) */
#define FPS_IDLE        1                       /* Has used FP instructions, but not in the middle of one */
#define FPS_BUSY        2                       /* In the middle of an instruction */

struct fp_istate
  {
    unsigned char version;
    unsigned char fsize;
    unsigned short res1;
    union FPF_u2
      {
        struct
          {
            unsigned short ccr;
            unsigned short res2;
            unsigned int xop[3];
            unsigned int opreg;
            unsigned int biu;
          } idle881;

        struct
          {
            unsigned short ccr;
            unsigned short res2;
            unsigned int iregs1[8];
            unsigned int xop[3];
            unsigned int opreg;
            unsigned int biu;
          } idle882;

        struct
          {
            unsigned int iregs[45];
          } busy881;

        struct
          {
            unsigned int iregs[53];
          } busy882;

      } u;
  };

#endif /* _machine_frame_ */
