/*
 * Mach Operating System
 * Copyright (c) 1992 Carnegie Mellon University
 * All Rights Reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie Mellon
 * the rights to redistribute these changes.
 */
#ifndef _machine_pcb_
#define _machine_pcb_

#include <machine/frame.h>

/*
 * 68881 floating-point state.
 */
struct fpu_state
  {
    struct fp_status status;          /* user-visible FPU state */
    struct fp_istate istate;          /* internal FPU state */
  };

static inline void
fp_save(struct fpu_state *state)
  {
    asm volatile("fsave %0@; tstb %0@; beq 0f; fmovem fp0-fp7,%0@-; fmovem fpcr/fpsr" /*"/fpiar"*/ ",%0@-; 0:"
      : : "a" (&state->istate));
  }

static inline void
fp_restore(struct fpu_state *state)
  {
    asm volatile("tstb %0@; beq 0f; fmovem %1@+,fpcr/fpsr" /*"/fpiar"*/ "; fmovem %1@,fp0-fp7; 0: frestore %0@"
      : : "a" (&state->istate), "a" (&state->status));
  }

/*
 * User state is allocated lots of space, since it varies in
 * size and moves.  Its TOP is fixed, not its bottom.
 */
#define USER_STATE_EXTRA        64      /* 16 words for stack switching */

#define USER_STATE_SIZE (USER_STATE_EXTRA + sizeof(struct frame))

struct pcb {
        unsigned char           flag;           /* user - trace flags */
        unsigned char           discard_frame;  /* user - discard bus-error
                                                          info on exit */
        unsigned short          fault_type;     /* user - fault type */
        vm_offset_t             fault_addr;     /* user - faulting address */
        struct frame *          user_regs;      /* pointer to user state */
        char                    user_state[USER_STATE_SIZE];
                                                /* user state pushed here */
        struct fpu_state        fpu_state;      /* FPU state, if in use */
};
typedef struct pcb      *pcb_t;

/*
 * Bits in flag byte.
 */
#define AST_STEP        0x40            /* force a single step */
#define TRACE_USER      0x20            /* user has requested tracing */
#define TRACE_AST       0x10            /* AST has requested tracing */
#define TRACE_PENDING   0x08            /* user trace caught in supervisor
                                           mode */
#define TRACE_KDB       0x04            /* trace is for KDB (single step
                                           supervisor mode) */
#define AST_CLR         (AST_STEP|TRACE_USER|TRACE_AST|TRACE_PENDING)



/*
 * State saved on kernel context switch.
 */
struct kernel_state {
        int             d2;
        int             d3;
        int             d4;
        int             d5;
        int             d6;
        int             d7;
        int             a2;
        int             a3;
        int             a4;
        int             a5;
        int             a6;
        int             sp;
};

#define STACK_KS(stack) \
        ((struct kernel_state *)((stack) + KERNEL_STACK_SIZE) - 1)


#endif _machine_pcb_
