/*
 * Mach Operating System
 * Copyright (c) 1991,1990,1989 Carnegie Mellon University
 * All Rights Reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie Mellon
 * the rights to redistribute these changes.
 */

#include <mach/std_types.h>
#include <mach/machine.h>

char version[] = "1.00";

char *root_name = "blah";

int boothowto = 0;

vm_size_t mem_size;

extern int page_size, page_shift;

extern char start[];
char *etext_ptr;

volatile void
ami_init(int phys_stext,int phys_etext,int phys_size,int chip_mem_size)
{
  int phys_end = phys_stext + phys_size;

  page_size = PAGE_SIZE;
  page_shift = PAGE_SHIFT;

  etext_ptr = start + phys_etext - phys_stext;
  mem_size = phys_size;
  vm_set_page_size();
  pmap_bootstrap(phys_stext,&phys_etext,&phys_end);
  trap_init();
  setup_main();
}

volatile void
halt_all_cpus(void)
{
  asm volatile("movew #0x2700,sr; movel #0x003f810f,sp@-; " /*pmove sp@,tt0;*/ "illegal; jmp 2");
}

volatile void
halt_cpu(void)
{
  asm volatile("illegal");
  asm volatile("stop #0x2700");
}

void
machine_init(void)
{
  machine_slot[0].is_cpu = 1;
  /* FIXME: type, subtype */
}

void
startrtclock(void)
{
}

void
resettodr(void)
{
}


#include <device/io_req.h>
#include <device/device_types.h>

static char console_buf[8192];
static char *console_ptr = console_buf;

int
cnputc(int c)
{
  *console_ptr++ = c;
  if(console_ptr >= console_buf+8192)
    {
      bcopy(console_buf+4096,console_buf,4096);
      bzero(console_buf+4096,4096);
      console_ptr = console_buf+4096;
    }
  return(0);
}

int
cngetc(void)
{
  asm("illegal");
  return('\n');
}

io_return_t
cnopen(int dev_number,int mode,io_req_t ior)
{
  return(D_SUCCESS);
}

void
cnclose(int dev_number)
{
}

io_return_t
cnwrite(int dev_number,io_req_t ior)
{
  char *p = ior->io_data;
  int size = ior->io_count;

  while(size--)
    cnputc(*p++);
}

io_return_t
cnread(int dev_number,io_req_t ior)
{
  return(D_INVALID_OPERATION);
}

io_return_t
cngetstat(int dev_number,int flavor,dev_status_t status,unsigned int status_count)
{
  return(D_INVALID_OPERATION);
}

io_return_t
cnsetstat(int dev_number,int flavor,dev_status_t status,unsigned int status_count)
{
  return(D_INVALID_OPERATION);
}

int timemmap(int dev_number,vm_offset_t ofs,vm_prot_t prot)
{
  return(0); /* ??? */
}


#include <mach/mach_interface.h>
#include <mach/thread_status.h>

/*
int
ex_get_header(struct file *fp,struct loader_info *lp)
{
  int result;
  vm_size_t resid;

  result = read_file(fp,0,lp,sizeof(*lp),&resid);
  if (result)
    return (result);
  if (resid)
    return (EX_NOT_EXECUTABLE);
  return(0);
}

char *
set_regs(mach_port_t user_task, mach_port_t user_thread, struct loader_info *lp, int arg_size)
{
  struct thread_state_regs regs;

  asm("illegal");
  regs.sp = VM_MAX_ADDRESS-arg_size-4;
  regs.pc = lp->text_start;
  thread_set_state(user_thread,THREAD_STATE_REGS,&regs,sizeof(regs)/sizeof(int));
  return(VM_MAX_ADDRESS-arg_size);
}
*/
